(in-package :xml-psychiatrist)

(defclass tag-matcher (matcher)
  ((child-matchers :initarg :child-matchers
		   :accessor child-matchers)
   (attribute-matchers :initarg :attribute-matchers
		       :accessor attribute-matchers)
   (name :initarg :name
	 :accessor name))
  (:documentation "A matcher that matches a tag"))

(defmethod match ((matcher tag-matcher) node)
  (with-slots (attribute-matchers child-matchers name) matcher
     (and (name-matches-p (safe-node-name node) name)
	  ;; Make sure that every attribute is one of the listed ones.
	  (every-node-is-matched-p attribute-matchers (node-attrs node)
				   :parent-node node
				   :node-type :attribute)
	  ;; Make sure that all required attributes are found.
	  (every-required-matcher-matches-something-p attribute-matchers (node-attrs node)
						      :parent-node node
						      :node-type :attribute)
	  ;; Make sure that every child is one of the listed ones.
	  (every-node-is-matched-p child-matchers (node-children node)
				   :parent-node node
				   :node-type :child)
	  ;; Make sure that all required children are found.
	  (every-required-matcher-matches-something-p child-matchers (node-children node)
						      :parent-node node
						      :node-type :child))))

(defmacro base-tag (name attributes children how-many)
  "Return matcher for a tag named `name', with given `attributes', whose `children' are specified in the body. Please note that
if the body is to contain PCDATA, this must be declared explicitly by placing `(pcdata)' somewhere in the list of children.

`name' can be either a string, a cons of (name . namespace), or a function"
  `(make-instance 'tag-matcher
		  :child-matchers (list ,@children)
		  :attribute-matchers (list ,@attributes)
		  :how-many ,how-many
		  :name ,name))

(defmacro tag (name attributes &body children)
  "A tag matcher that matches one tag"
  `(base-tag ,name ,attributes ,children :one))

(defmacro tag* (name attributes &body children)
  "A tag matcher that matches zero or more tags"
  `(base-tag ,name ,attributes ,children :zero-or-more))

(defmacro tag+ (name attributes &body children)
  "A tag matcher that matches one or more tags"
  `(base-tag ,name ,attributes ,children :one-or-more))

(defmacro tag? (name attributes &body children)
  "An optional tag matcher"
  `(base-tag ,name ,attributes ,children :zero-or-one))

(defun every-node-is-matched-p (matchers nodes &key parent-node (node-type :attribute))
  "Is every node matched by the matcher?"
  (dolist (node nodes)
    (unless (some-matcher-matches-p node matchers)
      (xml-validation-error "In ~S tag, the ~:[child tag~;attribute~] ~S is not allowed here."
			    (safe-node-name parent-node)
			    (case node-type
				  (:attribute t)
				  (:child nil)
				  (t (error "Bad node-type: ~S" node-type)))
			    (safe-node-name node))))
  (every #'identity 
	 (mapcar 
	  (lambda (node)
	    (some-matcher-matches-p node matchers))
	  nodes)))

(defun matcher-matches-enough-p (matcher nodes)
  "Does `matcher' match enough items in `nodes'?"
  (let ((number-of-matches (matcher-matches-count matcher nodes)))
    (ecase (how-many matcher)
      (:one (= number-of-matches 1))
      (:one-or-more (>= number-of-matches 1))
      (:zero-or-more (>= number-of-matches 0))
      (:zero-or-one (or (= number-of-matches 0)
			(= number-of-matches 1))))))

(defun every-required-matcher-matches-something-p (matchers nodes &key parent-node (node-type :attribute))
  "Does every matcher which is required (required slot is t) match one or more of the nodes?"
  (dolist (matcher matchers)
    (unless (matcher-matches-enough-p matcher nodes)
      (xml-validation-error "In ~S tag, the required ~:[child tag~;attribute~] ~S was not found."
			    (safe-node-name parent-node)
			    (case node-type
				  (:attribute t)
				  (:child nil)
				  (t (error "Bad node-type: ~S" node-type)))
			    (name matcher))))
  (every #'identity
	 (mapcar
	  (lambda (matcher)	   ; Does the matcher match enough?
	    (matcher-matches-enough-p matcher nodes))
	  matchers)))

(defun some-matcher-matches-p (item matchers)
  "Is item matched by one or more matchers?"
  (some #'identity
	(loop for matcher
	      in matchers
	      collect (match matcher item))))

(defun matcher-matches-count (matcher list)
  "Return the number of items in `list' that `matcher' matches"
  (count-if #'identity
	    (loop for item
		  in list
		  collect (match matcher item))))